@@ -1,21 +1,23 @@
 # Imports from stdlib:
 import time
-import socketserver
+import socket
 import struct
 import threading
 import logging
+import multiprocessing
 
 # Internal imports:
 from . import modbus_datastore
 
-socketserver.TCPServer.allow_reuse_address = True
 
 logger = logging.getLogger("modbus_server_logger")
-formatter = logging.Formatter("%(levelname)s: %(message)s")
+logger.setLevel(logging.DEBUG)
+formatter = logging.Formatter("%(levelname)-10s: %(message)s")
 streamhandler = logging.StreamHandler()
 streamhandler.setLevel(logging.DEBUG)
 streamhandler.setFormatter(formatter)
 logger.addHandler(streamhandler)
+logger.propagate = False  # prevent double logging
 
 
 def pack_bools_to_bytes(bool_list):
@@ -44,12 +46,20 @@ def build_error_response(header_items, exception_code):
     return response
 
 
-class TCPHandler(socketserver.BaseRequestHandler):
-    def handle(self):
-        """Callback Function for requests to the TCP Server:"""
+def handle_requests(s, addr, datastore):
+
+    while True:
 
         # Recv max 255 bytes, the maximal length for a Modbus frame:
-        self.data = self.request.recv(255).strip()
+        data = s.recv(256)
+        if not data:
+            break
+        if len(data) < 12:
+            logger.error(
+                f'Received less than 12 bytes, Header: {struct.unpack("!HHHBB", data[:8])}'
+            )
+            s.close()
+            break
 
         ## Extract Header + Function Code:
         # Transaction ID:   (2 Bytes)   Identifies the request-response-pair, is echoed in the response
@@ -58,12 +68,16 @@ class TCPHandler(socketserver.BaseRequestHandler):
         # Unit ID:          (1 Byte)    "Slave ID", inner identifier to route to different units (typically 0)
         # Function Code:    (1 Byte)    1,2,3,4,5,6,15,16,43: Read/Write input/register etc.
         try:
-            transaction_id, protocol, length, unit_id, function_code = struct.unpack(
-                "!HHHBB", self.data[:8]
-            )
+            (
+                transaction_id,
+                protocol,
+                length,
+                unit_id,
+                function_code,
+            ) = struct.unpack("!HHHBB", data[:8])
         except struct.error:
-            logger.error(f"Received incompatible bytes {self.data}")
-            return
+            logger.error(f"Received incompatible bytes {data}")
+            continue
 
         # Pack header items into a tuple which can more easily be passed around:
         header_items = (transaction_id, protocol, length, unit_id, function_code)
@@ -72,12 +86,12 @@ class TCPHandler(socketserver.BaseRequestHandler):
         if function_code not in (1, 2, 3, 4):
             # Respond with exception 01 - Illegal Function:
             response = build_error_response(header_items, exception_code=1)
-            self.request.sendall(response)
-            return None
+            s.sendall(response)
+            continue
 
         if function_code in (1, 2, 3, 4):  # -> The 4 'Read' Function Codes
-            first_address = struct.unpack("!H", self.data[8:10])[0]
-            number_of_registers = struct.unpack("!H", self.data[10:12])[0]
+            first_address = struct.unpack("!H", data[8:10])[0]
+            number_of_registers = struct.unpack("!H", data[10:12])[0]
 
         function_code_map = {
             1: "coils",
@@ -93,40 +107,37 @@ class TCPHandler(socketserver.BaseRequestHandler):
         if object_reference in ("coils", "discrete_inputs"):
             if number_of_registers < 1 or number_of_registers > 2000:
                 response = build_error_response(header_items, exception_code=3)
-                self.request.sendall(response)
-                return None
+                s.sendall(response)
+                continue
 
         if object_reference in ("input_registers", "holding_registers"):
             if number_of_registers < 1 or number_of_registers > 125:
                 response = build_error_response(header_items, exception_code=3)
-                self.request.sendall(response)
-                return None
+                s.sendall(response)
+                continue
 
         ## Read addresses from datastore
         ## =============================
 
         try:
-            # self.server is a reference to the socketserver.ThreadingTCPServer instance defined below:
-            data = self.server.datastore.read(
-                object_reference, first_address, number_of_registers
-            )
+            data = datastore.read(object_reference, first_address, number_of_registers)
         except KeyError:
             # Address not in datastore -> Respond with exception 02 - Illegal Data Address:
             response = build_error_response(header_items, exception_code=2)
             logger.warning(
-                f"Request from {self.client_address[0]} for {object_reference}:{first_address} -> Modbus Error 2"
+                f"Request from {addr[0]} for {object_reference}:{first_address} -> Modbus Error 2: Illegal Data Address"
             )
-            self.request.sendall(response)
-            return
+            s.sendall(response)
+            continue
         except:
             logger.error(
-                f"Request from {self.client_address[0]} for {object_reference}:{first_address} -> Modbus Error 4"
+                f"Request from {addr[0]} for {object_reference}:{first_address} -> Modbus Error 4"
             )
             raise
             # Other Error -> Respond with exception 04 - Slave Device Failure:
             response = build_error_response(header_items, exception_code=4)
-            self.request.sendall(response)
-            return
+            s.sendall(response)
+            continue
 
         ## Compose response
         ## ================
@@ -155,49 +166,62 @@ class TCPHandler(socketserver.BaseRequestHandler):
         response = struct.pack(f"!HHHBBB", *response_items) + data_bytes
 
         logger.debug(
-            f"Request from {self.client_address[0]} for {object_reference}:{first_address}+{number_of_registers} -> Response {data_bytes}"
+            f"Request from {addr[0]} for {object_reference}:{first_address}+{number_of_registers} -> Response {data_bytes}"
         )
-
-        self.request.sendall(response)
+        s.sendall(response)
 
 
 class Server:
     def __init__(
-        self, host="localhost", port=502, daemon=False, datastore=None, loglevel='INFO', autostart=False
+        self,
+        host="localhost",
+        port=502,
+        datastore=None,
+        loglevel="INFO",
+        autostart=False,
     ):
+        streamhandler.setLevel(loglevel)
         self.host = host
         self.port = port
         if datastore is None:
             self.datastore = modbus_datastore.DictDatastore()
         else:
             self.datastore = datastore
-        self.daemon = daemon
-        self.tcp_server = None
-        streamhandler.setLevel(loglevel)
+        self.server_thread = None
+        self.stop_server = False
         if autostart:
             self.start()
 
-    def _server_thread(self):
-        try:
-            self.tcp_server = socketserver.TCPServer((self.host, self.port), TCPHandler)
-        except OSError:
-            raise OSError(f"Could not start TCP server on port {self.port}")
-        self.tcp_server.allow_reuse_address = True
-        self.tcp_server.datastore = self.datastore
-        self.tcp_server.serve_forever()
-
     def start(self):
-        self.server_thread = threading.Thread(target=self._server_thread)
-        # daemon == True -> Shutdown the thread when the main program exits
-        self.server_thread.daemon = self.daemon
+        self.server_thread = threading.Thread(target=self._start_accepting)
+        # self.server_thread.daemon = True
         self.server_thread.start()
         logger.info("Modbus Server started")
 
+    def _start_accepting(self):
+        while not self.stop_server:
+            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
+                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+                s.bind((self.host, self.port))
+                s.listen(20)
+                con, addr = s.accept()
+                logger.debug(f"Connected to {addr}")
+                handling_thread = threading.Thread(
+                    target=handle_requests, args=(con, addr, self.datastore)
+                )
+                handling_thread.daemon = True
+                handling_thread.start()
+
     def stop(self):
-        if self.tcp_server:
-            self.tcp_server.shutdown()
-            self.tcp_server.server_close()
-            self.tcp_server = None
+        self.stop_server = True
+        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
+            try:
+                s.connect((self.host, self.port))
+            except ConnectionRefusedError:
+                pass
+        if self.server_thread:
+            self.server_thread.join(timeout=2)
+            self.server_thread = None
         logger.info("Modbus Server stopped")
 
     def dump_datastore(self):
